"""Handles filesystem interactions for neural network input data.

###############################################################################
# Filesystem Connector for Neural Network Input Data
# Copyright (c) 2017-2020 Joshua Burt
###############################################################################
"""

###############################################################################
# Dependencies
###############################################################################
import os
import fnmatch
import struct
import array
import logging
from collections import ValuesView
from typing import Dict, Any, List, Union

import numpy
from PIL import Image
from numpy import ndarray

from ..config import DiceboxConfig
from ..utils import lucky


class FileSystemConnector(object):
    """File System Connector Class"""
    dataset_index = None
    data_directory = None
    category_map = None
    pixel_cache = {}

    config = None

    def __init__(self, data_directory, disable_data_indexing=False, config_file='./dicebox.config',
                 lonestar_model_file='./dicebox.lonestar.json'):
        if self.config is None:
            self.config = DiceboxConfig(config_file=config_file,
                                        lonestar_model_file=lonestar_model_file)

        if self.data_directory is None:
            self.data_directory = os.path.normpath(data_directory)
            logging.info('data directory: (%s)', self.data_directory)

        if disable_data_indexing is False:
            if self.dataset_index is None:
                self.dataset_index = self.get_data_set()
                logging.debug('dataset_index')
                logging.debug(self.dataset_index)

            if self.category_map is None:
                self.category_map = self.get_data_set_categories()
                logging.debug('CATEGORY_MAP')
                logging.debug(self.category_map)
        else:
            logging.info('File System Connector Data Indexing Disabled.')

    def get_batch_list(self, batch_size: int):
        """For a given batch size, returns a random selection of indices

        :param batch_size: integer value
        :return: array of indices in the batch size (each index appearing only once).
        """
        output = []
        set_size: int = len(self.dataset_index)
        value_list = list(self.dataset_index.values())
        if batch_size > set_size:
            raise Exception('Max batch size: %s, but %s was specified!' % (set_size, batch_size))

        set_indices = []
        for i in range(0, set_size):
            set_indices.append(i)

        output_list = []
        while len(output_list) < batch_size:
            index: int = int(round((float(ord(struct.unpack('c', os.urandom(1))[0])) / 255) * (len(set_indices) - 1)))
            output_list.append(set_indices[index])
            set_indices.remove(set_indices[index])

        for i in output_list:
            output.append(value_list[i])
        return output

    def get_batch(self, batch_size: int, noise: int = 0) -> List[Union[list, List[ndarray]]]:
        """For a given batch size and noise level, returns a dictionary of data and labels.

        :param batch_size: integer
        :param noise: floating point number
        :return: dictionary of image data and labels
        """
        image_data = []
        image_labels = []

        category_map: Dict[Any, int] = self.category_map
        batch_list: [] = self.get_batch_list(batch_size)

        # build file path
        for i in range(0, batch_size):
            item = batch_list[i]
            filename = "%s/%s/%s" % (self.data_directory, item[1], item[0])
            logging.debug("(%s)(%s)", category_map[item[1]], filename)
            # logging.info("  natural category label: (%s)" % item[1])
            # logging.info("  neural network category label: (%i)" % category_map[item[1]])
            cat_one_hot = numpy.zeros(len(category_map))
            cat_one_hot[int(category_map[item[1]])] = 1
            image_labels.append(cat_one_hot)
            # image_labels.append(category_map[item[1]])

            # Help prevent over-fitting, and allow for new
            # sensory data to enter the cache, even when a cache
            # hit would occur.
            if filename in self.pixel_cache and not lucky(noise):
                del self.pixel_cache[filename]

            # use pixel cache if possible
            # [k,v] (filename, pixeldata)
            if filename in self.pixel_cache:
                # found in cache
                pixel_data = self.pixel_cache[filename]
                logging.debug("loaded cached pixel data for (%s)", filename)
            else:
                pixel_data = self.process_image(filename, noise)
                self.pixel_cache[filename] = pixel_data  # add to cache
                logging.debug("cached pixel data for (%s)", filename)

            image_data.append(pixel_data)
        return [image_data, image_labels]

    def process_image(self, filename, noise=0):
        """For a given filename, and noise level, will return a numpy array of pixel data.

        :param filename:
        :param noise:
        :return:
        """
        pixel_data = array.array('B')

        local_image = Image.open(filename).convert('L')  # Load as gray

        original_width, original_height = local_image.size
        # original_size = original_width, original_height
        logging.debug("original size: (%i, %i)", original_width, original_height)

        # See if there is noise
        if float(noise) > float(ord(struct.unpack('c', os.urandom(1))[0])) / 255:
            #  # then we introduce noise
            rotation_angle = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255 * 10
            if float(ord(struct.unpack('c', os.urandom(1))[0])) / 255 > 0.5:
                rotation_angle = rotation_angle * -1
            logging.debug("Rotating image %f", rotation_angle)
            local_image = local_image.rotate(rotation_angle)
            logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Perform a random scale
        if float(noise) > float(ord(struct.unpack('c', os.urandom(1))[0])) / 255:
            # random_scale = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
            random_scale = 1.5 * float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
            while random_scale < 0.7:
                random_scale = 1.5 * float(ord(struct.unpack('c', os.urandom(1))[0])) / 255

            logging.debug("Scaling image %f", random_scale)
            width, height = local_image.size
            new_width = int(width * random_scale)
            new_height = int(height * random_scale)
            if new_width > 0 and new_height > 0:
                new_size = int(width * random_scale), int(height * random_scale)
                # logging.info(new_size)
                local_image = local_image.resize(new_size)
                # local_image = local_image.resize(new_size, Image.ANTIALIAS)
                logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Crop Image If Required
        logging.debug('Crop image if required')

        # Now ensure we are the same dimensions as when we started
        new_width, new_height = local_image.size
        # if new_width > original_width or new_height > original_height:
        new_middle_x = float(new_width) / 2
        new_middle_y = float(new_height) / 2
        left = int(new_middle_x - float(original_width) / 2)
        upper = int(new_middle_y - float(original_height) / 2)
        right = int(new_middle_x + float(original_width) / 2)
        lower = int(new_middle_y + float(original_height) / 2)
        logging.debug("left: %i", left)
        logging.debug("upper: %i", upper)
        logging.debug("right: %i", right)
        logging.debug("lower: %i", lower)

        local_image = local_image.crop((left, upper, right, lower))
        logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # Ensure the input will match in input tensor
        # local_image = local_image.resize((original_width, original_height), Image.ANTIALIAS)
        local_image = local_image.resize((self.config.IMAGE_WIDTH, self.config.IMAGE_HEIGHT), Image.ANTIALIAS)
        logging.debug("new size: (%i, %i)", local_image.size[0], local_image.size[1])

        # dump to file for manual review
        # filename = datetime.now().strftime('transform_%Y-%m-%d_%H_%M_%S_%f.png')
        # local_image.save("./tmp/%s" % filename)

        pixel = local_image.load()

        width, height = local_image.size

        for x in range(0, width):
            for y in range(0, height):
                pixel_value = pixel[x, y]
                # logging.info("(%f)" % float(noise))
                # chance = float(ord(struct.unpack('c', os.urandom(1))[0])) / 255
                # logging.info("(%f)" % chance)
                # if float(noise) >= float(chance):
                #  # logging.info("  adding pixel noise (%f/%f)" % (float(noise), chance)) # add noise
                #  pixel_value = int(ord(struct.unpack('c', os.urandom(1))[0]))
                pixel_data.append(pixel_value)

        data = numpy.frombuffer(pixel_data, dtype=numpy.uint8)
        return data

    def get_data_set_categories(self):
        """Returns the dataset index as a dictionary of categories

        :return: category map as a dictionary
        """
        natural_categories = []
        category_map = {}
        value_list = self.dataset_index.values()
        # print(self.dataset_index)
        # print(value_list)
        for item in value_list:
            # logging.info(item)
            # logging.info("natural category label: (%s)" % item[1])
            natural_categories.append(item[1])
        natural_categories = sorted(set(natural_categories))
        cat_index = 0
        for cat in natural_categories:
            # logging.info("%i: %s" % (cat_index, cat))
            category_map[cat] = cat_index
            cat_index += 1
        # logging.info(category_map)
        return category_map

    def get_data_set(self) -> dict:
        """
        Returns a dictionary of [k:filename, v:array of filename and category] for the entire data set.

        When we move this to python 3, there are much better libraries to handle this.  Checkout PurePath..
        """
        data_set = {}
        for root, _, filenames in os.walk(self.data_directory):
            for filename in fnmatch.filter(filenames, '*.png'):
                new_entry = str(os.path.join(root, filename))
                new_entry = new_entry.replace('%s/' % self.data_directory, '')
                new_entry = os.path.normpath(new_entry)
                category, filename = new_entry.split('/')
                # logging.info("category: (%s), filename: (%s)" % (category, filename))
                data_set[new_entry] = [filename, category]
        # logging.info(data_set)
        return data_set



# Derived from https://github.com/harvitronix/neural-network-genetic-algorithm
# Derived source copyright: Matt Harvey, 2017, Derived source license: The MIT License
# See docs/Matt Harvey.LICENSE.txt

"""
Class that holds a genetic algorithm for evolving a network.

Credit:
    A lot of those code was originally inspired by:
    http://lethain.com/genetic-algorithms-cool-name-damn-simple/
"""
from functools import reduce
from operator import add
import copy

from .config import DiceboxConfig
from .dicebox_network import DiceboxNetwork
from .utils import lucky, random_index, random_index_between, random, random_strict


class EvolutionaryOptimizer:
    """Class that implements genetic algorithm for MLP optimization."""

    config_file = None
    lonestar_model_file = None
    config = None

    def __init__(self,
                 retain=0.4,
                 random_select=0.1,
                 mutate_chance=0.2,
                 config_file='./dicebox.config',
                 lonestar_model_file='./dicebox.lonestar.json'):
        if self.config_file is None:
            self.config_file = config_file

        if self.lonestar_model_file is None:
            self.lonestar_model_file = lonestar_model_file

        if self.config is None:
            self.config = DiceboxConfig(config_file=config_file,
                                        lonestar_model_file=lonestar_model_file)

        """Create an optimizer.

        Args:
            retain (float): Percentage of population to retain after
                each generation
            random_select (float): Probability of a rejected network
                remaining in the population
            mutate_chance (float): Probability a network will be
                randomly mutated

        """
        self.mutate_chance = mutate_chance
        self.random_select = random_select
        self.retain = retain

    def create_population(self, count):
        """Create a population of random networks.

        Args:
            count (int): Number of networks to generate, aka the
                size of the population

        Returns:
            (list): Population of network objects

        """
        pop = []
        for _ in range(0, count):
            # Create a random network.
            network = DiceboxNetwork(config_file=self.config_file,
                                     lonestar_model_file=self.lonestar_model_file)
            network.create_random()

            # Add the network to our population.
            pop.append(network)

        return pop

    @staticmethod
    def fitness(network):
        """Return the accuracy, which is our fitness function."""
        return network.accuracy

    def grade(self, pop):
        """Find average fitness for a population.

        Args:
            pop (list): The population of networks

        Returns:
            (float): The average accuracy of the population

        """
        summed = reduce(add, (self.fitness(network) for network in pop))
        return summed / float((len(pop)))

    def breed(self, mother, father):
        """Make two children as parts of their parents.

        Args:
            mother (dict): Network parameters
            father (dict): Network parameters

        Returns:
            (list): Two network objects

        """
        children = []
        for _ in range(2):
            child = DiceboxNetwork(config_file=self.config_file,
                                     lonestar_model_file=self.lonestar_model_file)
            if child.network is None:
                child.network = {}
            if 'layers' not in child.network:
                child.network['layers'] = []

            # Set unchange-ables
            child.network['input_shape'] = child.config.INPUT_SHAPE
            child.network['output_size'] = child.config.NB_CLASSES

            # Pick which parent's optimization function is passed on to offspring
            if lucky(0.5):
                # logging.debug("child.network['optimizer'] = mother(%s)", mother.network['optimizer'])
                child.network['optimizer'] = mother.network['optimizer']
            else:
                # logging.debug("child.network['optimizer'] = father(%s)", father.network['optimizer'])
                child.network['optimizer'] = father.network['optimizer']

            # Determine the number of layers
            if lucky(0.5):
                # logging.debug("child layer length = mother(%s)", len(mother.network['layers']))
                layer_count = len(mother.network['layers'])
            else:
                # logging.debug("child layer length = father(%s)", len(father.network['layers']))
                layer_count = len(father.network['layers'])

            for layer_index in range(0, layer_count):
                # logging.debug("layer (%s/%s)", layer_index, layer_count)
                # Pick which parent's layer is passed on to the offspring
                if lucky(0.5):
                    if layer_index < len(mother.network['layers']):
                        child.network['layers'].append(mother.network['layers'][layer_index])
                    elif layer_index < len(father.network['layers']):
                        child.network['layers'].append(father.network['layers'][layer_index])
                    else:
                        raise Exception('impossible breeding event occurred!')
                else:
                    if layer_index < len(father.network['layers']):
                        child.network['layers'].append(father.network['layers'][layer_index])
                    elif layer_index < len(mother.network['layers']):
                        child.network['layers'].append(mother.network['layers'][layer_index])
                    else:
                        raise Exception('impossible breeding event occurred!')

            child.model = child.compile_model(child.network)
            children.append(child)
        return children

    def mutate(self, individual: DiceboxNetwork) -> DiceboxNetwork:
        # we will be performing a deepcopy on the incoming object.
        # It looks like Keras Sequencials no longer support this.
        # so we need to ensure we remove any compiled models on
        # the inbound object before proceeding.
        individual.model = {}

        # mutations = 0
        local_noise = self.mutate_chance
        # logging.debug("***************************************************")
        clone = copy.deepcopy(individual)
        # see if the optimizer is mutated
        if lucky(local_noise):
            # yep..  Select an optimizer
            # logging.debug("optimizer = (%s)", clone.network['optimizer'])
            clone.network['optimizer'] = clone.select_random_optimizer()
            # mutations += 1
            # logging.debug("optimizer = (%s)", clone.network['optimizer'])

        # Determine the number of layers..
        layer_count = len(clone.network['layers'])

        # now mess around within the layers
        for index in range(1, layer_count):
            # see if the layer is mutated
            if lucky(local_noise):
                # then change the layer type
                # how does this affect the weights, etc? :/
                # logging.debug("layer = (%s)", layer)
                clone.network['layers'][index - 1] = clone.build_random_layer()
                # mutations += 1
                # logging.debug("layer = (%s)", layer)
            else:
                layer = clone.network['layers'][index - 1]

                # keep checking the individual layer attributes
                if layer['type'] == 'dropout':
                    if lucky(local_noise):
                        # mutate the dropout rate
                        # logging.debug("rate = (%s)", layer['rate'])
                        layer['rate'] = random_strict()
                        # mutations += 1
                        # logging.debug("rate = (%s)", layer['rate'])
                elif layer['type'] == 'dense':
                    if lucky(local_noise):
                        # mutate the layer size
                        # logging.debug('Mutating layer size')
                        # logging.debug("size = (%s)", layer['size'])
                        layer['size'] = random_index_between(clone.config.TAXONOMY['min_neurons'],
                                                                     clone.config.TAXONOMY['max_neurons'])
                        # mutations += 1
                        # logging.debug("size = (%s)", layer['size'])
                    if lucky(local_noise):
                        # mutate activation function
                        # logging.debug("activation = (%s)", layer['activation'])
                        activation_index = random_index(len(clone.config.TAXONOMY['activation']))
                        layer['activation'] = clone.config.TAXONOMY['activation'][activation_index - 1]
                        # mutations += 1
                        # logging.debug("activation = (%s)", layer['activation'])
                else:
                    # logging.debug('Unknown layer type')
                    raise Exception('Not yet implemented!')
        # logging.debug("mutations: (%s)", mutations)
        # logging.debug("***************************************************")
        return clone

    def evolve(self, pop):
        """Evolve a population of networks.

        Args:
            pop (list): A list of network parameters

        Returns:
            (list): The evolved population of networks

        """
        # Get scores for each network.
        graded = [(self.fitness(network), network) for network in pop]

        # Sort on the scores.
        graded = [x[1] for x in sorted(graded, key=lambda x: x[0], reverse=True)]

        # Get the number we want to keep for the next gen.
        retain_length = int(len(graded)*self.retain)

        # The parents are every network we want to keep.
        parents = copy.deepcopy(graded[:retain_length])

        # For those we aren't keeping, randomly keep some anyway.
        for individual in graded[retain_length:]:
            if self.random_select > random():
                parents.append(copy.deepcopy(individual))

        # Randomly mutate some of the networks we're keeping.
        for individual in parents:
            if lucky(self.mutate_chance):
                individual = self.mutate(individual)

        # Now find out how many spots we have left to fill.
        parents_length = len(parents)
        desired_length = len(pop) - parents_length
        children = []

        # Add children, which are bred from two remaining networks.
        while len(children) < desired_length:

            # Get a random mom and dad.
            male_index = random_index_between(0, parents_length - 1)
            female_index = random_index_between(0, parents_length - 1)

            # Assuming they aren't the same network...
            if male_index != female_index:
                male = parents[male_index]
                female = parents[female_index]

                # Breed them.
                babies = self.breed(male, female)

                # Add the children one at a time.
                for baby in babies:
                    # Don't grow larger than desired length.
                    if len(children) < desired_length:
                        children.append(baby)

        parents.extend(children)

        return parents



"""Allow over-riding the defaults with non-secure ENV variables, or secure docker secrets

###############################################################################
# Local Config File Handler
# Copyright (c) 2017-2020 Joshua Burt
###############################################################################
"""

###############################################################################
# Dependencies
###############################################################################
import os
import json
from .base_config import BaseConfig


class DiceboxConfig(object):
    
    def __init__(self, config_file='./dicebox.config', lonestar_model_file='./dicebox.lonestar.json'):
        self.dc = BaseConfig(config_file=config_file, lonestar_model_file=lonestar_model_file)

        ###############################################################################
        # Data Set Options
        ###############################################################################

        # Load user defined config
        self.DATASET = self.dc.DATASET
        if 'DATASET' in os.environ:
            self.DATASET = os.environ['DATASET']

        self.DICEBOX_COMPLIANT_DATASET = self.dc.DICEBOX_COMPLIANT_DATASET
        if 'DICEBOX_COMPLIANT_DATASET' in os.environ:
            self.DICEBOX_COMPLIANT_DATASET = os.environ['DICEBOX_COMPLIANT_DATASET']

        self.NB_CLASSES = self.dc.NB_CLASSES
        if 'NB_CLASSES' in os.environ:
            self.NB_CLASSES = int(os.environ['NB_CLASSES'])

        self.IMAGE_WIDTH = self.dc.IMAGE_WIDTH
        if 'IMAGE_WIDTH' in os.environ:
            self.IMAGE_WIDTH = int(os.environ['IMAGE_WIDTH'])

        self.IMAGE_HEIGHT = self.dc.IMAGE_HEIGHT
        if 'IMAGE_HEIGHT' in os.environ:
            self.IMAGE_HEIGHT = int(os.environ['IMAGE_HEIGHT'])

        self.DATA_BASE_DIRECTORY = self.dc.DATA_BASE_DIRECTORY
        if 'DATA_BASE_DIRECTORY' in os.environ:
            self.DATA_BASE_DIRECTORY = os.environ['DATA_BASE_DIRECTORY']


        ###############################################################################
        # Build Calculated Configs
        ###############################################################################
        self.NETWORK_NAME = "%s_%ix%i" % (self.DATASET, self.IMAGE_WIDTH, self.IMAGE_HEIGHT)
        if 'NETWORK_NAME' in os.environ:
            self.NETWORK_NAME = os.environ['NETWORK_NAME']

        self.INPUT_SHAPE = (self.IMAGE_WIDTH * self.IMAGE_HEIGHT,)
        if 'INPUT_SHAPE' in os.environ:
            self.INPUT_SHAPE = os.environ['INPUT_SHAPE']

        self.DATA_DIRECTORY = "%s/%s/data/" % (self.DATA_BASE_DIRECTORY, self.NETWORK_NAME)
        if 'DATA_DIRECTORY' in os.environ:
            self.DATA_DIRECTORY = os.environ['DATA_DIRECTORY']


        ###############################################################################
        # Neural Network Taxonomy Options
        ###############################################################################
        self.MIN_NEURONS = self.dc.MIN_NEURONS
        if 'MIN_NEURONS' in os.environ:
            self.MIN_NEURONS = os.environ['MIN_NEURONS']

        self.MAX_NEURONS = self.dc.MAX_NEURONS
        if 'MAX_NEURONS'  in os.environ:
            self.MAX_NEURONS = os.environ['MAX_NEURONS']

        self.MIN_LAYERS = self.dc.MIN_LAYERS
        if 'MIN_LAYERS'  in os.environ:
            self.MIN_LAYERS = os.environ['MIN_LAYERS']

        self.MAX_LAYERS = self.dc.MAX_LAYERS
        if 'MAX_LAYERS'  in os.environ:
            self.MAX_LAYERS = os.environ['MAX_LAYERS']

        self.LAYER_TYPES = self.dc.LAYER_TYPES
        if 'LAYER_TYPES'  in os.environ:
            self.LAYER_TYPES = os.environ['LAYER_TYPES']

        self.ACTIVATION = self.dc.ACTIVATION
        if 'ACTIVATION' in os.environ:
            self.ACTIVATION = os.environ['ACTIVATION']

        self.OPTIMIZER = self.dc.OPTIMIZER
        if 'OPTIMIZER' in os.environ:
            self.OPTIMIZER = os.environ['OPTIMIZER']

        self.TAXONOMY = {
            'min_neurons': self.MIN_NEURONS,
            'max_neurons': self.MAX_NEURONS,
            'min_layers': self.MIN_LAYERS,
            'max_layers': self.MAX_LAYERS,
            'layer_types': json.loads(self.LAYER_TYPES),
            'activation': json.loads(self.ACTIVATION),
            'optimizer': json.loads(self.OPTIMIZER)
        }


        ###############################################################################
        # Lonestar Options
        ###############################################################################
        self.LONESTAR_DICEBOX_MODEL = self.dc.LONESTAR_DICEBOX_MODEL
        if 'LONESTAR_DICEBOX_MODEL' in os.environ:
            self.LONESTAR_DICEBOX_MODEL = json.loads(os.environ['LONESTAR_DICEBOX_MODEL'])


        ###############################################################################
        # Evolution Options
        ###############################################################################
        self.EPOCHS = self.dc.EPOCHS
        if 'EPOCHS' in os.environ:
            self.EPOCHS = int(os.environ['EPOCHS'])

        self.GENERATIONS = self.dc.GENERATIONS  # Number of times to evole the population.
        if 'GENERATIONS' in os.environ:
            self.GENERATIONS = int(os.environ['GENERATIONS'])

        self.POPULATION = self.dc.POPULATION  # Number of networks in each generation.
        if 'POPULATION' in os.environ:
            self.POPULATION = int(os.environ['POPULATION'])

        self.NOISE = self.dc.NOISE
        if 'NOISE' in os.environ:
            self.NOISE = float(os.environ['NOISE'])


        ###############################################################################
        # Training Options / Settings for the 1920x1080 dataset
        ###############################################################################
        self.BATCH_SIZE = self.dc.BATCH_SIZE
        if 'BATCH_SIZE' in os.environ:
            self.BATCH_SIZE = int(os.environ['BATCH_SIZE'])

        self.TRAIN_BATCH_SIZE = self.dc.TRAIN_BATCH_SIZE
        if 'TRAIN_BATCH_SIZE' in os.environ:
            self.TRAIN_BATCH_SIZE = int(os.environ['TRAIN_BATCH_SIZE'])

        self.TEST_BATCH_SIZE = self.dc.TEST_BATCH_SIZE
        if 'TEST_BATCH_SIZE' in os.environ:
            self.TEST_BATCH_SIZE = int(os.environ['TEST_BATCH_SIZE'])

        self.LOAD_BEST_WEIGHTS_ON_START = self.dc.LOAD_BEST_WEIGHTS_ON_START
        if 'LOAD_BEST_WEIGHTS_ON_START' in os.environ:
            if os.environ['LOAD_BEST_WEIGHTS_ON_START'] == 'False':
                self.LOAD_BEST_WEIGHTS_ON_START = False


        ###############################################################################
        # Direcrtory Options
        ###############################################################################
        self.LOGS_DIR = self.dc.LOGS_DIR
        if 'LOGS_DIR' in os.environ:
            self.LOGS_DIR = os.environ['LOGS_DIR']

        self.WEIGHTS_DIR = self.dc.WEIGHTS_DIR
        if 'WEIGHTS_DIR' in os.environ:
            self.WEIGHTS_DIR = os.environ['WEIGHTS_DIR']

        self.TMP_DIR = self.dc.TMP_DIR
        if 'TMP_DIR' in os.environ:
            self.MP_DIR = os.environ['TMP_DIR']


        ###############################################################################
        # Server Options
        ###############################################################################
        self.API_ACCESS_KEY = self.dc.API_ACCESS_KEY
        if 'API_ACCESS_KEY' in os.environ:
            self.API_ACCESS_KEY = os.environ['API_ACCESS_KEY']

        self.API_VERSION = self.dc.API_VERSION
        if 'API_VERSION' in os.environ:
            self.API_VERSION = os.environ['API_VERSION']

        self.LISTENING_HOST = self.dc.LISTENING_HOST
        if 'LISTENING_HOST' in os.environ:
            self.LISTENING_HOST = os.environ['LISTENING_HOST']

        self.FLASK_DEBUG = self.dc.FLASK_DEBUG
        if 'FLASK_DEBUG' in os.environ:
            if os.environ['FLASK_DEBUG'] == 'True':
                self.FLASK_DEBUG = True

        self.MODEL_WEIGHTS_FILENAME = self.dc.MODEL_WEIGHTS_FILENAME
        if 'MODEL_WEIGHTS_FILENAME' in os.environ:
            self.MODEL_WEIGHTS_FILENAME = os.environ['MODEL_WEIGHTS_FILENAME']


        ###############################################################################
        # Sensory Service Options
        ###############################################################################
        self.SENSORY_SERVER = self.dc.SENSORY_SERVER
        if 'SENSORY_SERVER' in os.environ:
            self.SENSORY_SERVER = os.environ['SENSORY_SERVER']

        self.SENSORY_PORT = self.dc.SENSORY_PORT
        if 'SENSORY_PORT' in os.environ:
            self.SENSORY_PORT = os.environ['SENSORY_PORT']

        self.SENSORY_URI = self.dc.SENSORY_URI
        if 'SENSORY_URI' in os.environ:
            self.SENSORY_URI = os.environ['SENSORY_URI']

        self.SENSORY_SERVICE_RABBITMQ_EXCHANGE = self.dc.SENSORY_SERVICE_RABBITMQ_EXCHANGE
        if 'SENSORY_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_EXCHANGE = os.environ['SENSORY_SERVICE_RABBITMQ_EXCHANGE']

        self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY = self.dc.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY
        if 'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY = os.environ['SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_ROUTING_KEY']

        self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE = self.dc.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE
        if 'SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE = os.environ['SENSORY_SERVICE_RABBITMQ_BATCH_REQUEST_TASK_QUEUE']

        self.SENSORY_SERVICE_RABBITMQ_URI = self.dc.SENSORY_SERVICE_RABBITMQ_URI
        if 'SENSORY_SERVICE_RABBITMQ_URI' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_URI = os.environ['SENSORY_SERVICE_RABBITMQ_URI']

        self.SENSORY_SERVICE_RABBITMQ_USERNAME = self.dc.SENSORY_SERVICE_RABBITMQ_USERNAME
        if 'SENSORY_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_USERNAME = os.environ['SENSORY_SERVICE_RABBITMQ_USERNAME']

        self.SENSORY_SERVICE_RABBITMQ_PASSWORD = self.dc.SENSORY_SERVICE_RABBITMQ_PASSWORD
        if 'SENSORY_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_PASSWORD = os.environ['SENSORY_SERVICE_RABBITMQ_PASSWORD']

        self.SENSORY_SERVICE_RABBITMQ_SERVER = self.dc.SENSORY_SERVICE_RABBITMQ_SERVER
        if 'SENSORY_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_SERVER = os.environ['SENSORY_SERVICE_RABBITMQ_SERVER']

        self.SENSORY_SERVICE_RABBITMQ_PORT = self.dc.SENSORY_SERVICE_RABBITMQ_PORT
        if 'SENSORY_SERVICE_RABBITMQ_PORT' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_PORT = os.environ['SENSORY_SERVICE_RABBITMQ_PORT']

        self.SENSORY_SERVICE_RABBITMQ_VHOST = self.dc.SENSORY_SERVICE_RABBITMQ_VHOST
        if 'SENSORY_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.SENSORY_SERVICE_RABBITMQ_VHOST = os.environ['SENSORY_SERVICE_RABBITMQ_VHOST']

        self.SENSORY_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.SENSORY_SERVICE_RABBITMQ_URI,
            self.SENSORY_SERVICE_RABBITMQ_USERNAME,
            self.SENSORY_SERVICE_RABBITMQ_PASSWORD,
            self.SENSORY_SERVICE_RABBITMQ_SERVER,
            self.SENSORY_SERVICE_RABBITMQ_PORT,
            self.SENSORY_SERVICE_RABBITMQ_VHOST
        )
        self.SENSORY_SERVICE_SHARD_SIZE = self.dc.SENSORY_SERVICE_SHARD_SIZE
        if 'SENSORY_SERVICE_SHARD_SIZE' in os.environ:
            self.SENSORY_SERVICE_SHARD_SIZE = int(os.environ['SENSORY_SERVICE_SHARD_SIZE'])



        ###############################################################################
        # Training Service Options
        ###############################################################################
        self.TRAINING_SERVICE_RABBITMQ_EXCHANGE = self.dc.TRAINING_SERVICE_RABBITMQ_EXCHANGE
        if 'TRAINING_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_EXCHANGE = os.environ['TRAINING_SERVICE_RABBITMQ_EXCHANGE']

        self.TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = self.dc.TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY
        if 'TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = os.environ['TRAINING_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY']

        self.TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = self.dc.TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE
        if 'TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = os.environ['TRAINING_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE']

        self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_VHOST = self.dc.TRAINING_SERVICE_RABBITMQ_RABBITMQ_VHOST
        if 'TRAINING_SERVICE_RABBITMQ_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_VHOST = os.environ['TRAINING_SERVICE_RABBITMQ_RABBITMQ_VHOST']

        self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI = self.dc.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI
        if 'TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI = os.environ['TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI']

        self.TRAINING_SERVICE_RABBITMQ_USERNAME = self.dc.TRAINING_SERVICE_RABBITMQ_USERNAME
        if 'TRAINING_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_USERNAME = os.environ['TRAINING_SERVICE_RABBITMQ_USERNAME']

        self.TRAINING_SERVICE_RABBITMQ_PASSWORD = self.dc.TRAINING_SERVICE_RABBITMQ_PASSWORD
        if 'TRAINING_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_PASSWORD = os.environ['TRAINING_SERVICE_RABBITMQ_PASSWORD']

        self.TRAINING_SERVICE_RABBITMQ_SERVER = self.dc.TRAINING_SERVICE_RABBITMQ_SERVER
        if 'TRAINING_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_SERVER = os.environ['TRAINING_SERVICE_RABBITMQ_SERVER']

        self.TRAINING_SERVICE_RABBITMQ_PORT = self.dc.TRAINING_SERVICE_RABBITMQ_PORT
        if 'TRAINING_SERVICE_RABBITMQ_PORT' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_PORT = os.environ['TRAINING_SERVICE_RABBITMQ_PORT']

        self.TRAINING_SERVICE_RABBITMQ_VHOST = self.dc.TRAINING_SERVICE_RABBITMQ_VHOST
        if 'TRAINING_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_SERVICE_RABBITMQ_VHOST = os.environ['TRAINING_SERVICE_RABBITMQ_VHOST']

        self.TRAINING_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_SERVICE_RABBITMQ_RABBITMQ_URI,
            self.TRAINING_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_SERVICE_RABBITMQ_PORT,
            self.TRAINING_SERVICE_RABBITMQ_VHOST
        )


        ###############################################################################
        # Training Processor Options
        ###############################################################################
        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_EXCHANGE']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAINING_REQUEST_ROUTING_KEY']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_TRAIN_REQUEST_TASK_QUEUE']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_VHOST = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_VHOST
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_VHOST = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_VHOST']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST = self.dc.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST
        if 'TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST' in os.environ:
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST = os.environ['TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST']

        self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_URL = "%s%s:%s@%s:%s/%s" % (
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_RABBITMQ_URI,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_USERNAME,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PASSWORD,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_SERVER,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_PORT,
            self.TRAINING_PROCESSOR_SERVICE_RABBITMQ_VHOST
        )


        ###############################################################################
        # Client Options
        ###############################################################################
        self.CLASSIFICATION_SERVER = self.dc.CLASSIFICATION_SERVER
        if 'CLASSIFICATION_SERVER' in os.environ:
            self.CLASSIFICATION_SERVER = os.environ['CLASSIFICATION_SERVER']

        self.SERVER_PORT = self.dc.SERVER_PORT
        if 'SERVER_PORT' in os.environ:
            self.SERVER_PORT = int(os.environ['SERVER_PORT'])

        self.SERVER_URI = self.dc.SERVER_URI
        if 'SERVER_URI' in os.environ:
            self.SERVER_URI = os.environ['SERVER_URI']



##############################################################################
# Derived from https://github.com/harvitronix/neural-network-genetic-algorithm
# Derived source copyright: Matt Harvey, 2017, Derived source license: The MIT License
# See docs/Matt Harvey.LICENSE.txt
##############################################################################

"""Class that represents the network to be evolved."""
from typing import Union, Dict, Any

from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.callbacks import EarlyStopping
# from tensorflow.keras.callbacks import ModelCheckpoint
import logging
import numpy
from datetime import datetime
import os
import json

from .config import DiceboxConfig
from .connectors import FileSystemConnector, SensoryServiceConnector
from .utils import random_index, random_index_between, random, random_strict


class DiceboxNetwork:
    """Represent a network and let us operate on it.

    Currently only works for an MLP.
    """
    fsc: FileSystemConnector = None  # file system connector
    ssc: SensoryServiceConnector = None  # sensory service connector

    config: DiceboxConfig = None

    # Helper: Early stopping.
    early_stopper = EarlyStopping(patience=25)

    ##############################################################################
    # Feature disabled until a flipper can be added and the filenames created safely.
    # Since this now runs in a container some additional considerations must be made.
    ##############################################################################
    # Checkpoint
    # filepath = "%s/weights-improvement-{epoch:02d}-{val_acc:.2f}.hdf5" % config.WEIGHTS_DIR
    # checkpoint = ModelCheckpoint(filepath, monitor='val_acc', verbose=1, save_best_only=True, mode='max')
    # callbacks_list = [early_stopper, checkpoint]

    callbacks_list = [early_stopper]

    def __init__(self,
                 nn_param_choices=None,
                 create_fcs: bool = True,
                 disable_data_indexing: bool = False,
                 config_file: str = './dicebox.config',
                 lonestar_model_file: str = './dicebox.lonestar.json'):
        if self.config is None:
            self.config = DiceboxConfig(config_file=config_file,
                                        lonestar_model_file=lonestar_model_file)

        """Initialize our network.

        Args:
            nn_param_choices (dict): Parameters for the network, includes:
                nb_neurons (list): [64, 128, 256]
                nb_layers (list): [1, 2, 3, 4]
                activation (list): ['relu', 'elu']
                optimizer (list): ['rmsprop', 'adam']
        """

        self.accuracy: float = 0.
        self.network = {}  # (the dicebox object)
        self.model: Sequential | None = None  # the compiled network.

        if self.fsc is None and create_fcs is True:
            logging.debug('creating a new fsc..')
            logging.info('self.config.DATA_DIRECTORY: (%s)' % self.config.DATA_DIRECTORY)
            self.fsc = FileSystemConnector(data_directory=self.config.DATA_DIRECTORY,
                                           disable_data_indexing=disable_data_indexing,
                                           config_file=config_file,
                                           lonestar_model_file=lonestar_model_file)

        if self.ssc is None:
            logging.debug('creating a new ssc..')
            self.ssc = SensoryServiceConnector(role='client',
                                               config_file=config_file,
                                               lonestar_model_file=lonestar_model_file)

    def create_random(self):
        self.network = {}
        self.network['layers'] = []

        # Set unchange-ables
        self.network['input_shape'] = self.config.INPUT_SHAPE
        self.network['output_size'] = self.config.NB_CLASSES

        # Select an optimizer
        optimizer_index: int = random_index(len(self.config.TAXONOMY['optimizer']))
        optimizer: str = self.config.TAXONOMY['optimizer'][optimizer_index - 1]
        self.network['optimizer'] = optimizer

        # Determine the number of layers..
        layer_count: int = random_index_between(self.config.TAXONOMY['min_layers'],
                                                self.config.TAXONOMY['max_layers'])
        for layer_index in range(1, layer_count):
            # add new random layer to the network
            self.network['layers'].append(self.build_random_layer())

    def select_random_optimizer(self) -> int:
        # Select an optimizer
        optimizer_index: int = random_index(len(self.config.TAXONOMY['optimizer']))
        return self.config.TAXONOMY['optimizer'][optimizer_index - 1]

    def build_random_layer(self) -> Dict[str, Union[str, float, int]]:
        # determine what the layer type will be
        layer_type_index = random_index(len(self.config.TAXONOMY['layer_types']))
        layer_type = self.config.TAXONOMY['layer_types'][layer_type_index - 1]

        random_layer = {}
        random_layer['type']: str = layer_type
        if layer_type == 'dropout':
            # get a dropout rate..
            random_layer['rate']: float = random_strict()
        else:
            # determine the size and activation function to use.
            random_layer['size']: int = random_index_between(self.config.TAXONOMY['min_neurons'],
                                                        self.config.TAXONOMY['max_neurons'])
            activation_index: int = random_index(len(self.config.TAXONOMY['activation']))
            random_layer['activation']: str = self.config.TAXONOMY['activation'][activation_index - 1]
        return random_layer

    def create_lonestar(self, create_model: bool = False, weights_filename: str = None) -> None:
        logging.debug('-' * 80)
        logging.debug('create_lonestar(create_model, weights_filename)')
        logging.debug(create_model)
        logging.debug(weights_filename)
        logging.debug('-' * 80)

        self.network: Any = self.config.LONESTAR_DICEBOX_MODEL
        logging.debug('-' * 80)
        logging.debug(self.network)
        logging.debug('-' * 80)

        if create_model is True:
            if self.model is None:
                logging.debug('compiling model')
                self.model: Sequential = DiceboxNetwork.compile_model(self.network)
                if weights_filename is not None:
                    logging.debug("loading weights file: (%s)" % weights_filename)
                    self.load_model(weights_filename)
            # else:
            #     logging.info('model already compiled, skipping.')

    def create_set(self, network: Any) -> None:
        self.network = network

    def train(self) -> None:
        if self.accuracy == 0.:
            self.accuracy = self.train_and_score(self.network)

    def train_and_save(self, dataset: Any) -> None:
        # if self.accuracy == 0.:
        logging.debug('-' * 80)
        logging.debug("train_and_save(dataset)")
        logging.debug("train_and_save(dataset=%s)" % dataset)
        logging.debug('-' * 80)
        self.accuracy = self.train_and_score_and_save(dataset)

    def print_network(self) -> None:
        """Print out a network."""
        logging.info(self.network)
        logging.info("Network accuracy: %.2f%%" % (self.accuracy * 100))

    def train_and_score(self, network):
        if self.config.DICEBOX_COMPLIANT_DATASET is True:
            x_train, x_test, y_train, y_test = self.get_dicebox_filesystem()
        else:
            raise Exception('Unknown dataset type!  Please define, or correct.')

        model = DiceboxNetwork.compile_model(network)

        logging.info('batch size (model.fit): %s' % self.config.BATCH_SIZE)

        logging.info('Fitting network:')
        logging.info(network)
        logging.info('compiled model:')
        logging.info(json.dumps(json.loads(model.to_json())))

        model.fit(x_train, y_train,
                  batch_size=self.config.BATCH_SIZE,
                  epochs=10000,  # using early stopping, so no real limit
                  verbose=1,
                  validation_data=(x_test, y_test),
                  callbacks=[self.early_stopper])

        score = model.evaluate(x_test, y_test, verbose=0)

        return score[1]  # 1 is accuracy. 0 is loss.

    @staticmethod
    def compile_model(dicebox_model):
        model = Sequential()

        layers = dicebox_model['layers']
        first_layer = False
        for layer in layers:
            # build and add layer
            if layer['type'] == 'dropout':
                # handle dropout
                model.add(Dropout(layer['rate']))
            else:
                neurons = layer['size']
                activation = layer['activation']

                if first_layer is False:
                    first_layer = True
                    model.add(Dense(neurons, activation=activation, input_shape=dicebox_model['input_shape']))
                else:
                    model.add(Dense(neurons, activation=activation))

        # add final output layer.
        model.add(Dense(dicebox_model['output_size'], activation='softmax'))

        model.compile(loss='categorical_crossentropy', optimizer=dicebox_model['optimizer'],
                      metrics=['accuracy'])

        return model

    def get_dicebox_filesystem(self):
        noise = self.config.NOISE
        test_batch_size = self.config.TEST_BATCH_SIZE
        train_batch_size = self.config.TRAIN_BATCH_SIZE

        logging.info('noise: %s' % noise)
        logging.info('train_batch_size: %s' % train_batch_size)
        logging.info('test_batch_size: %s' % test_batch_size)

        train_image_data, train_image_labels = self.fsc.get_batch(train_batch_size, noise=noise)
        # train_image_data, train_image_labels = Network.ssc.get_batch(train_batch_size, noise=noise)
        train_image_data = numpy.array(train_image_data)
        train_image_data = train_image_data.astype('float32')
        train_image_data /= 255
        train_image_labels = numpy.array(train_image_labels)

        test_image_data, test_image_labels = self.fsc.get_batch(test_batch_size, noise=noise)
        # test_image_data, test_image_labels = Network.ssc.get_batch(test_batch_size, noise=noise)
        test_image_data = numpy.array(test_image_data)
        test_image_data = test_image_data.astype('float32')
        test_image_data /= 255
        test_image_labels = numpy.array(test_image_labels)

        x_train = train_image_data
        x_test = test_image_data
        y_train = train_image_labels
        y_test = test_image_labels

        return x_train, x_test, y_train, y_test

    def get_dicebox_sensory_data(self):
        logging.debug('-' * 80)
        logging.debug('get_dicebox_sensory_data(self)')
        logging.debug('-' * 80)

        noise = self.config.NOISE
        train_batch_size = self.config.TRAIN_BATCH_SIZE
        test_batch_size = self.config.TEST_BATCH_SIZE

        try:
            # train_image_data, train_image_labels = Network.fsc.get_batch(train_batch_size, noise=noise)
            train_image_data, train_image_labels = self.ssc.get_batch(train_batch_size, noise=noise)

            logging.debug('-' * 80)
            logging.debug('train_image_data to numpy.array')
            # logging.debug(train_image_data)

            train_image_data = numpy.array(train_image_data)
            # logging.debug(train_image_data)

            logging.debug('train_image_data astype float32')
            train_image_data = train_image_data.astype('float32')
            # logging.debug(train_image_data)

            logging.debug('train_image_data /255')
            train_image_data /= 255
            # logging.debug(train_image_data)

            logging.debug('train_image_labels to numpy.array')
            train_image_labels = numpy.array(train_image_labels)
            # logging.debug(train_image_labels)
            logging.debug('-' * 80)
        except ValueError:
            logging.debug('Caught ValueError when processing training data.')
            logging.debug('failing out..')
            raise ValueError

        try:
            # test_image_data, test_image_labels = Network.fsc.get_batch(test_batch_size, noise=noise)
            test_image_data, test_image_labels = self.ssc.get_batch(test_batch_size, noise=noise)

            logging.debug('-' * 80)
            logging.debug('test_image_data to numpy.array')
            # logging.debug(test_image_data)

            test_image_data = numpy.array(test_image_data)
            # logging.debug(test_image_data)

            logging.debug('test_image_data astype float32')
            test_image_data = test_image_data.astype('float32')
            # logging.debug(test_image_data)

            logging.debug('test_image_data /255')
            test_image_data /= 255
            # logging.debug(test_image_data)

            logging.debug('test_image_labels to numpy.array')
            test_image_labels = numpy.array(test_image_labels)
            # logging.debug(test_image_labels)
        except ValueError:
            logging.debug('Caught ValueError when processing test data.')
            logging.debug('failing out..')
            raise ValueError

        x_train = train_image_data
        x_test = test_image_data
        y_train = train_image_labels
        y_test = test_image_labels
        return x_train, x_test, y_train, y_test

    def train_and_score_and_save(self, dataset):
        logging.debug('-' * 80)
        logging.debug("train_and_score_and_save(dataset)")
        logging.debug("train_and_score_and_save(dataset=%s)" % dataset)
        logging.debug('-' * 80)
        if self.config.DICEBOX_COMPLIANT_DATASET is True:
            logging.debug('-' * 80)
            logging.debug('loading sensory data..')
            logging.debug('-' * 80)
            x_train, x_test, y_train, y_test = self.get_dicebox_sensory_data()
            # nb_classes, batch_size, input_shape, x_train, x_test, y_train, y_test = self.get_dicebox_sensory_data()
            # nb_classes, batch_size, input_shape, x_train, x_test, y_train, y_test = self.get_dicebox_filesystem()
            logging.debug('-' * 80)
            logging.debug('Done!')
            logging.debug('-' * 80)
        else:
            # no support yet!
            logging.error('UNSUPPORTED dataset supplied to train_and_score_and_save')
            raise Exception('UNSUPPORTED dataset supplied to train_and_score_and_save')

        logging.debug('-' * 80)
        logging.debug('Compiling model if need be.')
        logging.debug('-' * 80)
        if self.model is None:
            self.model = DiceboxNetwork.compile_model(self.network)
        logging.debug('-' * 80)
        logging.debug('Done!')
        logging.debug('-' * 80)

        logging.debug('-' * 80)
        logging.debug('Fitting model.')
        logging.debug('-' * 80)
        self.model.fit(x_train, y_train,
                       batch_size=self.config.BATCH_SIZE,
                       epochs=10000,  # using early stopping, so no real limit
                       verbose=1,
                       validation_data=(x_test, y_test),
                       callbacks=self.callbacks_list)
        logging.debug('-' * 80)
        logging.debug('Done!')
        logging.debug('-' * 80)

        logging.debug('-' * 80)
        logging.debug('Scoring model.')
        logging.debug('-' * 80)
        score = self.model.evaluate(x_test, y_test, verbose=1)
        logging.debug('-' * 80)
        logging.debug('Done!')
        logging.debug('-' * 80)

        return score[1]  # 1 is accuracy. 0 is loss.

    def save_model(self, filename):
        if self.model is None:
            logging.error('no model! :(  compile the model first.')
            raise Exception('no model! :(  compile the model first.')
        logging.debug('loading weights file..')
        try:
            self.model.load_weights(str(filename))  # https://github.com/keras-team/keras/issues/11269
        except Exception as e:
            logging.error('Unable to load weights file.')
            logging.error(e)
            raise e

    def load_model(self, filename):
        if self.model is None:
            logging.error('no model! :(  compile the model first.')
            raise Exception('no model! :(  compile the model first.')
        logging.debug('loading weights file..')
        try:
            self.model.load_weights(str(filename))  # https://github.com/keras-team/keras/issues/11269
        except Exception as e:
            logging.error('Unable to load weights file.')
            logging.error(e)
            raise e
        raise Exception('Not yet implemented!')

    def classify(self, network_input):
        if self.config.DICEBOX_COMPLIANT_DATASET is True:
            x_test = self.get_dicebox_raw(network_input)
        else:
            logging.error("UNKNOWN DATASET (%s) passed to classify" % self.config.NETWORK_NAME)
            raise Exception("UNKNOWN DATASET (%s) passed to classify" % self.config.NETWORK_NAME)

        if self.model is None:
            logging.error('Unable to classify without a model. :(')
            raise Exception('Unable to classify without a model. :(')

        model_prediction = self.model.predict_classes(x_test, batch_size=1, verbose=0)
        logging.info(model_prediction)

        return model_prediction

    def get_dicebox_raw(self, raw_image_data):
        # ugh dump to file for the time being
        filename = "%s/%s" % (self.config.TMP_DIR, datetime.now().strftime('%Y-%m-%d_%H_%M_%S_%f.tmp.png'))
        with open(filename, 'wb') as f:
            f.write(raw_image_data)

        try:
            test_image_data = self.fsc.process_image(filename)
        except:
            logging.error('Exception caught processing image data.')
            raise Exception('Exception caught processing image data.')
        finally:
            os.remove(filename)

        test_image_data = numpy.array(test_image_data)
        test_image_data = test_image_data.astype('float32')
        test_image_data /= 255

        x_test = [test_image_data]
        x_test = numpy.array(x_test)

        return x_test
